from django.contrib import admin

from .models import OktaUser


class OktaUserAdmin(admin.ModelAdmin):
	'''User admin
	Customized user admin for Okta users.
	'''
	
	fieldsets = (
		('Basic Info', {'fields': ('login', 'firstName', 'lastName', 'email')}),
		('Names', {
			'classes'	: ('collapse',),
			'fields'	: ('displayName', 'honorificPrefix', 'middleName', 'honorificSuffix', 'nickName'),
			}),
		('Contact Info', {
			'classes'	: ('collapse',),
			'fields'	: ('primaryPhone', 'mobilePhone', 'profileUrl', 'secondEmail'),
			}),
		('Organization', {
			'classes'	: ('collapse',),
			'fields'	: ('organization', 'division', 'department', 'title', 'userType', 'employeeNumber', 'costCenter', 'managerId', 'manager'),
			}),
		('Addresses', {
			'classes'	: ('collapse',),
			'fields'	: ('streetAddress', 'city', 'state', 'zipCode', 'countryCode', 'postalAddress'),
			}),
		('International', {
			'classes'	: ('collapse',),
			'fields'	: ('locale', 'preferredLanguage', 'timezone'),
			}),
		('Groups', {
			'classes'	: ('collapse',),
			'fields'	: ('groups',),
			}),
		('Permissions', {'fields': ('is_staff', 'is_superuser', 'is_active')}),
	)
	list_display = ('login', 'firstName', 'lastName', 'email', 'is_staff', 'is_superuser', 'is_active')
	list_filter = ('is_active', 'userType', 'organization', 'division', 'department', 'title', 'managerId')
	ordering = ('login',)
	search_fields = ('login', 'firstName', 'lastName', 'email')
	
	def save_model(self, request, obj, form, change):
		
		if change:
			return super().save_model(request, obj, form, change)
		else:
			return type(obj).objects.create_user(**form.cleaned_data)


admin.site.register(OktaUser, OktaUserAdmin)



from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.core.validators import RegexValidator
from django.db import models
from django.utils import timezone as timezone_utils
from django.utils.translation import gettext_lazy as _

from .managers import OktaUserManager

validator_5_to_100 = RegexValidator(regex = '^.{5,100}$', message = '5 <= value length <= 100')
nullable_5_to_100 = RegexValidator(regex = '(^.{5,100}$)|(ˆ$)', message = '5 <= value length <= 100 (or nothing)')

class AbstractOktaUser(AbstractBaseUser, PermissionsMixin):
	'''Default Okta profile
	Based on the official documentation as of (10/2021) https://developer.okta.com/docs/reference/api/users/#default-profile-properties
	'''
	
	login = models.CharField(primary_key = True, max_length = 100, validators=[validator_5_to_100], verbose_name = _('login'), help_text = _('unique identifier for the user (username)'))
	email = models.EmailField(blank = False, validators=[validator_5_to_100], verbose_name = _('email'), help_text = _('primary email address of user'))
	secondEmail = models.EmailField(blank = True, validators=[nullable_5_to_100], verbose_name = _('second email'), help_text = _('secondary email address of user typically used for account recovery'))
	firstName = models.CharField(blank = False, max_length = 50, verbose_name = _('first name'), help_text = _('given name of the user (givenName)'))
	lastName = models.CharField(blank = False, max_length = 50, verbose_name = _('last name'), help_text = _('family name of the user (familyName)'))
	middleName = models.CharField(blank = True, max_length = 50, verbose_name = _('middle name'), help_text = _('middle name(s) of the user'))
	honorificPrefix = models.CharField(blank = True, max_length = 50, verbose_name = _('honorific prefix'), help_text = _('honorific prefix(es) of the user, or title in most Western languages'))
	honorificSuffix = models.CharField(blank = True, max_length = 50, verbose_name = _('honorific suffix'), help_text = _('honorific suffix(es) of the user'))
	title = models.CharField(blank = True, max_length = 100, verbose_name = _('title'), help_text = _('user\'s title, such as "Vice President'))
	displayName = models.CharField(blank = True, max_length = 250, verbose_name = _('display name'), help_text = _('name of the user, suitable for display to end users'))
	nickName = models.CharField(blank = True, max_length = 50, verbose_name = _('nickname'), help_text = _('casual way to address the user in real life'))
	profileUrl = models.URLField(blank = True, verbose_name = _('profile URL'), help_text = _("url of user's online profile (e.g. a web page)"))
	primaryPhone = models.CharField(blank = True, max_length = 100, verbose_name = _('primary phone'), help_text = _('primary phone number of user such as home number'))
	mobilePhone = models.CharField(blank = True, max_length = 100, verbose_name = _('mobile phone'), help_text = _('mobile phone number of user'))
	streetAddress = models.CharField(blank = True, max_length = 100, verbose_name = _('street address'), help_text = _("full street address component of user's address"))
	city = models.CharField(blank = True, max_length = 100, verbose_name = _('city'), help_text = _("city or locality component of user's address (locality)"))
	state = models.CharField(blank = True, max_length = 100, verbose_name = _('state'), help_text = _("state or region component of user's address (region)"))
	zipCode = models.CharField(blank = True, max_length = 100, verbose_name = _('ZIP code'), help_text = _("zipcode or postal code component of user's address (postalCode)"))
	countryCode = models.CharField(blank = True, max_length = 2, verbose_name = _('country code'), help_text = _("country name component of user's address (country)"))
	postalAddress = models.CharField(blank = True, max_length = 100, verbose_name = _('postal address'), help_text = _("mailing address component of user's address"))
	preferredLanguage = models.CharField(blank = True, max_length = 100, verbose_name = _('preferred language'), help_text = _("user's preferred written or spoken languages"))
	locale = models.CharField(blank = True, max_length = 5, verbose_name = _('locale'), help_text = _("user's default location for purposes of localizing items such as currency, date time format, numerical representations, etc."))
	timezone = models.CharField(blank = True, max_length = 100, verbose_name = _('time zone'), help_text = _("user's time zone"))
	userType = models.CharField(blank = True, max_length = 100, verbose_name = _('user type'), help_text = _('used to describe the organization to user relationship such as "Employee" or "Contractor"'))
	employeeNumber = models.CharField(blank = True, max_length = 100, verbose_name = _('employee number'), help_text = _('organization or company assigned unique identifier for the user'))
	costCenter = models.CharField(blank = True, max_length = 100, verbose_name = _('cost center'), help_text = _('name of a cost center assigned to user'))
	organization = models.CharField(blank = True, max_length = 100, verbose_name = _('organization'), help_text = _("name of user's organization"))
	division = models.CharField(blank = True, max_length = 100, verbose_name = _('division'), help_text = _("name of user's division"))
	department = models.CharField(blank = True, max_length = 100, verbose_name = _('department'), help_text = _("name of user's department"))
	managerId = models.CharField(blank = True, max_length = 100, verbose_name = _('manager ID'), help_text = _("id of a user's manager"))
	manager = models.CharField(blank = True, max_length = 100, verbose_name = _('manager'), help_text = _("displayName of the user's manager"))
	
	is_staff = models.BooleanField(default = False, verbose_name = _("staff status"), help_text=_("Designates whether the user can log into this admin site."))
	is_active = models.BooleanField(default = True, verbose_name = _("active"), help_text = _('Designates whether this user should be treated as active. \nUnselect this instead of deleting accounts.'))
	date_joined = models.DateTimeField(verbose_name = _("date joined"), default = timezone_utils.now)
	
	objects = OktaUserManager()
	
	EMAIL_FIELD = 'email'
	USERNAME_FIELD = 'login'
	REQUIRED_FIELDS = ['email', 'firstName', 'lastName']
	
	class Meta:
		verbose_name = _('okta user')
		verbose_name_plural = _('okta users')
		abstract = True
	
	def __str__(self):
		return self.login
	
	@classmethod
	def _attributes_from_okta_profile(cls, okta_profile):
		attributes = {}
		for field in cls._meta.fields:
			if hasattr(okta_profile, field.name):
				okta_attr = getattr(okta_profile, field.name)
				if (okta_attr is None) or not len(okta_attr):
					continue
				if isinstance(field, models.fields.DateTimeField):
					okta_attr = datetime.datetime.fromisoformat(okta_attr.rstrip('Z'))
				attributes[field.name] = okta_attr
		return attributes

	@classmethod
	def from_okta_profile(cls, okta_profile):
		return cls(**cls._attributes_from_okta_profile(okta_profile))

	def get_full_name(self):
		if self.displayName:
			return self.displayName
		else:
			components = ('honorificPrefix', 'firstName', 'middleName', 'lastName', 'honorificSuffix')
			return ' '.join([getattr(self, component) for component in components if getattr(self, component)])

	def get_short_name(self):
		if self.nickName:
			return self.nickName
		else:
			return self.firstName

	def update_from_okta_profile(self, okta_profile):
		for field_name, field_value in self._attributes_from_okta_profile(okta_profile).items():
			setattr(self, field_name, field_value)
		return self


class OktaUser(AbstractOktaUser):
	'''Django user model
	Alternate user model for Django based on Okta profiles.
	'''
	
	class Meta(AbstractOktaUser.Meta):
		swappable = 'AUTH_USER_MODEL'



from django.contrib.auth.base_user import BaseUserManager


class OktaUserManager(BaseUserManager):
	'''Okta user manager
	Custom user manager to work with the Okta user model.
	'''
	
	use_in_migrations = True
	
	def create_user(self, login, email, firstName, lastName, password = None, **other_fields):
		'''Create a local user
		Create and save a user with the provided details.
		'''
		
		try:
			email = self.normalize_email(email)
		except Exception:
			pass
		
		user = self.model(login = login, email = email, firstName = firstName, lastName = lastName, **other_fields)
		user.set_unusable_password()
		user.save()
		
		return user

	def create_superuser(self, login, email, firstName, lastName, password = None, **other_fields):
		'''Create a local superuser
		Create and save a superuser with the provided details, including a password.
		'''
			
		try:
			email = self.normalize_email(email)
		except Exception:
			pass
		
		other_fields.setdefault('is_staff', True)
		other_fields.setdefault('is_superuser', True)
		other_fields.setdefault('is_active', True)

		if other_fields.get('is_staff') is not True:
			raise ValueError('Superuser must have is_staff=True.')
		if other_fields.get('is_superuser') is not True:
			raise ValueError('Superuser must have is_superuser=True.')
		
		user = self.model(login = login, email = email, firstName = firstName, lastName = lastName, **other_fields)
		user.set_password(password)
		user.save()
		
		return user



from django.urls import path, include

from . import apps
from . import views

app_name = apps.OktaClientConfig.name

urlpatterns = [
	path('saml/', views.acs, name = 'acs'),
	path('login/', views.login_, name = 'login'),
	path('logout/', views.logout_, name = 'logout'),
]



#python

import json
import logging

import saml2
import saml2.client 
import saml2.config

from django.conf import settings
from django.contrib.auth import login, logout, get_user_model
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseServerError, JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.csrf import csrf_exempt

from . import apps
from . import models

LOGGER = logging.getLogger(__name__)


class SPConfig:
	
	class OktaConfig(dict):
		
		def __init__(self, request, django_settings = settings):
			
			try:
				okta_settings = django_settings.OKTA_CLIENT
			except AttributeError:
				raise ValueError('Missing OKTA_CLIENT section in Django settings')
				
			super().__init__()
			
			local_domain_url = okta_settings.get('ASSERTION_DOMAIN_URL', '{}://{}'.format(request.scheme, request.get_host()))	
			acs_url = ''.join((local_domain_url, reverse('{}:acs'.format(apps.OktaClientConfig.name))))
			
			if 'METADATA_LOCAL_FILE_PATH' in okta_settings:
				self['metadata'] = {'local' : okta_settings['METADATA_LOCAL_FILE_PATH']}
			else:
				self['metadata'] = {'remote' : [{'url' : okta_settings['METADATA_AUTO_CONF_URL']}]}
				
			self['entityid'] = okta_settings.get('ENTITY_ID', acs_url)
			self['service'] = {
				'sp': {
					'endpoints': {
						'assertion_consumer_service': [
							(acs_url, saml2.BINDING_HTTP_REDIRECT),
							(acs_url, saml2.BINDING_HTTP_POST)
						],
					},
					'allow_unsolicited': True,
					'authn_requests_signed': False,
					'logout_requests_signed': True,
					'want_assertions_signed': True,
					'want_response_signed': False,
				},
			}
			
			if 'NAME_ID_FORMAT' in okta_settings:
				self['service']['sp']['name_id_format'] = okta_settings['NAME_ID_FORMAT']
	
	def __new__(cls, request, django_config = settings):
		
		sp_config = saml2.config.SPConfig()
		sp_config.load(cls.OktaConfig(request, django_config))
		sp_config.allow_unknown_attributes = True
		return sp_config
	
	@classmethod
	def next_url(cls, request, django_config = settings):
		config = cls.OktaConfig(request, django_config)
		return request.GET.get('next', config.get('DEFAULT_NEXT_URL', '/'))


########## ----- Mixins ----- ##########


class OktaEventHookMixin:
	'''Okta Event Hook endpoint
	Follow Okta's way to do Okta Event Hooks, for your convenience.
	
	Your class based view, the one that inherits this mixin, should implement an *okta_"type_of_event"* method for each event type that this view will handle where "type_of_event" will be the event type reported by okta with the dots replaced by underscores. Ex: okta_user_session_start
	
	There's also the option to implement a fallback "okta_event" method that will be used if no specific method exists for an event type.
	
	All these methods should expect 3 parameters:
	- request: is the Django request object provided to the view
	- event: is the already parsed event that should be processed in this iteration
	- event_targets: is the list of targets in the event as a dictionary keyd by type (an error will be raised if overlapping types). It could be "None" if no targets are present on the event object.
	
	Ref: https://developer.okta.com/docs/concepts/event-hooks/
	'''
	
	def get(self, request):
		'''HTTP GET
		Only used to confirm that it follows Okta's convention.
		'''
		
		return JsonResponse({'verification' : request.headers.get('x-okta-verification-challenge','')})
	
	def post(self, request, request_json = None, overrider_method_name = ''):
		'''HTTP GET
		Regular Event Hook handling.
		'''
		
		if request_json is None:
			request_json = json.loads(request.body)
		not_implemented = False
		for event in request_json['data']['events']:
			method_name = 'okta_' + event['eventType'].replace('.', '_')
			if ('target' in event) and (event['target'] is not None):
				event_targets = {}
				for target in event['target']:
					if target['type'] in event_targets:
						raise RuntimeError('Okta event with overlapping targets: {}'.format(event['uuid']))
					event_targets[target['type']] = {key : value for key, value in target.items() if key not in ['type']}
			else:
				event_targets = None
			if overrider_method_name:
				getattr(self, overrider_method_name)(request, request_json, event, event_targets)
			elif hasattr(self, method_name):
				getattr(self, method_name)(request, request_json, event, event_targets)
			elif hasattr(self, 'okta_event'):
				getattr(self, 'okta_event')(request, request_json, event, event_targets)
			else:
				not_implemented = True
				LOGGER.warning('Okta event support not implemented: %s', event['eventType'])

		if not_implemented:
			return HttpResponse(status = 501)
		else:
			return HttpResponse(status = 204)


########## ----- Views ----- ##########


@csrf_exempt
def acs(request):
	
	next_url = request.session.get('next_url', SPConfig.next_url(request))
	saml_client = saml2.client.Saml2Client(config = SPConfig(request))
	
	response = request.POST.get('SAMLResponse', None)
	if response:
		LOGGER.debug('ACS received SAML response: %s', response)
	else:
		LOGGER.error('No POST request to ACS')
		return HttpResponse(status = 407)

	authn_response = saml_client.parse_authn_request_response(response, saml2.entity.BINDING_HTTP_POST)
	if authn_response is None:
		LOGGER.error('Unable to parse SAML response: %s', response)
		return HttpResponse(status = 400)
	else:
		LOGGER.debug('Parsed SAML response: %s', authn_response)
	
	login_id = authn_response.get_subject().text
	
	if False:
		#Implement API calls and pull user from Okta
		pass
	else:
		user_identity = authn_response.get_identity()
		if user_identity is None:
			LOGGER.error('Malformed SAML response (get_identity failed): %s', authn_response)
			return HttpResponse(status = 401)
		else:
			LOGGER.debug('Identity correctly extracted: %s', user_identity)
	
		defaults = {key : value[0] if isinstance(value, list) and (len(value) == 1) else value for key, value in user_identity.items() if key not in ['login']}
		defaults['is_active'] = True
		LOGGER.debug('Updating or creating user "%s" with: %s', login_id, defaults)
		target_user, created_flag = get_user_model().objects.update_or_create(defaults = defaults, login = login_id)
	
	LOGGER.info('Logging in "%s"', target_user)
	login(request, target_user)
	
	LOGGER.debug('Redirecting after login to "%s"', next_url)
	return HttpResponseRedirect(next_url)

def login_(request):
	
	LOGGER.debug('Logging in: %s', request)
	
	request.session['next_url'] = SPConfig.next_url(request)
	LOGGER.debug('Saved "next_url" into session: %s', request.session['next_url'])
	
	saml_client = saml2.client.Saml2Client(config = SPConfig(request))
	LOGGER.debug('Preparing authentication with: %s', saml_client)
	session_id, request_info = saml_client.prepare_for_authenticate()
	LOGGER.debug('Session id %s includes: %s', session_id, request_info)

	for key, value in request_info['headers']:
		if key == 'Location':
			LOGGER.debug('Found "Location" header. Redirecting to "%s"', value)
			return HttpResponseRedirect(value)
	
	LOGGER.error('The "Location" header was not found')
	return HttpResponseServerError()


def logout_(request):
	if request.user.is_authenticated:
		LOGGER.info('Logging out user: %s', request.user)
		logout(request)
	else:
		LOGGER.debug('No authenticated user. Ignoring logout request')
		
	next_url = request.session.get('next_url', SPConfig.next_url(request))
	LOGGER.debug('Redirecting after logout: %s', next_url)
	return HttpResponseRedirect(next_url)

@login_required
def index(request):
	return render(request, 'okta-client/index.html')
